;MICRO-READERBOARD ORNAMENT
;    DOT MATRIX VERSION
;
;LED dot matrix display: columns = anode, rows = cathode
;ATTiny2313 @ 500 kHz (internal R/C osc.)
;Fuses:  low = 0x42 (internal 4MHz, divide Clock by 8, BOD enabled)
;       high = 0x9D (BOD level 1.8V, preserve EEPROM through chip erase)
;
;PORTS:
;  PB0+1 = Column 1 (left)
;  PB2+3 = Column 2
;  PB4+5 = Column 3
;  PB6+7 = Column 4
;  PA0+1 = Column 5 (right)
;  PD0   = Row 1 (top)
;  PD1   = Row 2
;  PD2   = Row 3 / Push-Button / RS232 RxD (connect to TxD from PC)
;  PD3   = Row 4 / RS232 GND
;  PD4   = Row 5 / Push-Button
;  PD5   = Row 6
;  PD6   = Row 7 (bottom)
;
;VERSION HISTORY:
;  Version 0.1     2007-01-07   Arne Rossius
;  Version 0.8     2007-01-16   Arne Rossius
;  Version 0.8.1   2007-07-07   Arne Rossius
;    * fixed bug in charset (char 0x35 = '5')
;  Version 0.8.2   2008-12-14   Arne Rossius
;    * made changes for displays with column cathode (Kingbright TC*)
;  Version 0.8.3   2010-02-10   Arne Rossius
;    * added option to select column cathode or anode
;  Version 0.9   2010-02-14   Arne Rossius
;    * added support for optical data transfer (enter via push-button)
;  Version 0.9.1   2010-12-28   Arne Rossius
;    * removed extra delay in RS232 reception



.include "../../include/tn2313def.inc"

.def	zero = R0

.def	w1 = R3
.def	scrollspeed = R4
.def	delay = R5
.def	col1 = R6
.def	col2 = R7
.def	col3 = R8
.def	col4 = R9
.def	col5 = R10
.def	char1 = R11
.def	char2 = R12
.def	char3 = R13
.def	char4 = R14
.def	char5 = R15

.def	temp = R16
.def	temp2 = R17
.def	count = R18
.def	bright = R19
.def	white = R20
.def	bits = R21
.def	avglen = R22
.def	rxbyte = R23
.def	charlen = R24

;R26/27 used for buffer (RAM) pointer
;R28/29 used for string (EEPROM) pointer
;R30/31 used for charset (Flash) pointer

.equ	LINE_OFF = 0x00
.equ	LINE1 = 0x03
.equ	LINE2 = 0x0C
.equ	LINE3 = 0x30
.equ	LINE4 = 0xC0
.equ	LINE5_OFF = 0x04
.equ	LINE5 = 0x07

;===============================================================================

.dseg
	RAM_String:
		.byte 112
	RAM_Stack:
		.byte 16

.eseg
	.org 0x00
	.db 0xff, 8, "Hello World!    ", 0x00

;===============================================================================

.cseg
.org 0x000
	rjmp	start
.org OVF0addr ;Timer 0 overflow interrupt
	rjmp	t0ovf
	
;===============================================================================
	
charset:
	.include "../charset.asm"
	
;===============================================================================

t0ovf:
	;Timer 0 overflow interrupt (~30 Hz)
	push	temp
	push	temp2
	in	temp, SREG
	push	temp
	
	;decrement delay counter
	dec	delay
	brne	t0ovf_end
	mov	delay, scrollspeed
	
	;shift display left by one column
	lsr	char1
	rol	col1
	lsr	char2
	rol	col2
	lsr	char3
	rol	col3
	lsr	char4
	rol	col4
	lsr	char5
	rol	col5
	subi	charlen, 1
	brcc	t0ovf_end ;overflow -> next char
nextchar:
	;load next character
	rcall	eeprom_read
	inc	YL
	mov	ZL, temp
	cpi	ZL, 0x00 ;end marker -> restart string
	breq	restart
	subi	ZL, 0x20 ;charset starts at 0x20
	clr	ZH
	lsl	ZL ;offset: multiply ZH:ZL by 6
	rol	ZH
	mov	temp, ZL
	mov	temp2, ZH
	lsl	ZL
	rol	ZH
	add	ZL, temp
	adc	ZH, temp2
	ldi	temp, LOW(charset*2) ;add charset address
	ldi	temp2, HIGH(charset*2)
	add	ZL, temp
	adc	ZH, temp2
	lpm	charlen, Z+ ;load and invert character
	lpm	char1, Z+
	lpm	char2, Z+
	lpm	char3, Z+
	lpm	char4, Z+
	lpm	char5, Z+
	
t0ovf_end:
	;return
	pop	temp
	out	SREG, temp
	pop	temp2
	pop	temp
	reti
	
restart:
	;restart string
	ldi	YL, 0x02
	rjmp	nextchar

;===============================================================================
	
start:
	;stack pointer
	ldi	temp, RAMEND
	out	SPL, temp
	
	;ports
	ldi	temp, 0x03
	out	DDRA, temp
	ldi	temp, 0xFF
	out	DDRB, temp
	ldi	temp, 0x7B
	out	DDRD, temp
	ldi	temp, LINE5_OFF
	out	PORTA, temp
	ldi	temp, LINE_OFF
	out	PORTB, temp
	ldi	temp, 0xFF
	out	PORTD, temp
	
	;zero register
	clr	zero
	
	;check if RS232 cable is connected
	rcall	muxwait
	sbic	PIND, 2 ;PD2 low in spite of pullup -> RS232 connected
	rjmp	play
	rjmp	transfer

;===============================================================================

;---------------------------------------
;       OPTICAL TRANSFER MODE
;---------------------------------------

rx:
	cli
	
	;TODO: timeout (no change for too long)
	;discharge LEDs
	ldi	temp, 0x00
	out	PORTD, temp
	ldi	temp, 0x7F
	out	DDRD, temp
	ldi	temp, 0x04
	out	PORTA, temp
	ldi	temp, 0x03
	out	DDRA, temp
	ldi	temp, 0x00
	out	PORTB, temp
	ldi	temp, 0xFF
	out	DDRB, temp
	rcall	muxwait
	
	ldi	temp, 0x00
	out	DDRD, temp
	out	DDRA, temp
	ldi	temp, 0x30
	out	DDRB, temp
	
	;take 4 samples of white, calculate average
	clr	white
	clr	temp
	ldi	count, 4
rx_init_white_loop:
	rcall	rx_read
	add	white, bright
	brcc	PC+2
	inc	temp
	dec	count
	brne	rx_init_white_loop
	lsr	temp
	ror	white
	lsr	temp
	ror	white
	;use 5/4 of average white value as threshold
	mov	temp, white
	lsr	temp
	lsr	temp
	add	white, temp
	
	;wait until screen turns black (first bit)
	ldi	XL, RAM_String
	clr	bits
	clr	ZL
	clr	ZH
rx_init_white_wait:
	rcall	rx_read
	cp	bright, white
	brlo	rx_init_white_wait
	rjmp	rx_bit
	
rx_loop:
	;bit reception loop
	clr	temp
rx_white:
	;wait while screen is white
	rcall	rx_read
	add	temp, bright
	brcs	rx_done ;white for ~1 second: end of transmission
	cp	bright, white
	brlo	rx_white
rx_bit:
	;receive bit
	mov	temp, bright
	rcall	rx_read
	cp	bright, white
	brlo	rx_bit_done ;white: end of bit
	add	temp, bright
	sub	temp, white ;TODO: subtract white_val, not 5/4*white_val, also find out why this is necessary
rx_bit_done:
	;one bit received
	inc	bits
	cpi	bits, 9
	brlo	rx_header
	lsl	rxbyte
	cp	temp, avglen
	brlo	PC+2
	ori	rxbyte, 0x01
	mov	temp, bits
	andi	temp, 0x07
	cpi	temp, 7 ;7 bit = 1 byte (MSB isn't sent because it's always 0)
	brne	rx_loop
	;one byte completed
	subi	bits, 7
	andi	rxbyte, 0x7F
	cpi	XL, RAM_String+112 ;prevent buffer overflow
	brsh	PC+2
	st	X+, rxbyte
	rjmp	rx_loop
	
rx_header:
	;header bit received
	add	ZL, temp
	brcc	PC+2
	inc	ZH
	cpi	bits, 8
	brne	rx_loop
	;calculate average of 8 header bits (divide by 8)
	lsr	ZH
	ror	ZL
	lsr	ZH
	ror	ZL
	lsr	ZH
	ror	ZL
	brcc	PC+2 ;round up if fractional part of result is >= 1/2
	inc	ZL
	mov	avglen, ZL
	rjmp	rx_loop

rx_done:
	;reception complete: save received string to EEPROM
	mov	count, XL
	subi	count, 1 ;remove checksum byte from message
	cpi	count, RAM_String+2
	brlo	rx_empty
	;verify checksum
	ld	temp, -X
	clr	temp2
	ldi	XL, RAM_String
rx_checksum_loop:
	ld	rxbyte, X+
	add	temp2, rxbyte
	cp	XL, count
	brlo	rx_checksum_loop
	andi	temp2, 0x7F
	cp	temp, temp2
	brne	rx_checksum_error
	;copy message to EEPROM
	ldi	XL, RAM_String
	ldi	YL, 0x01 ;first byte is scrolling speed
rx_store_loop:
	ld	temp, X+
	rcall	eeprom_write
	inc	YL
	cp	XL, count
	brlo	rx_store_loop
	ldi	temp, 0x00
	rcall	eeprom_write
	rjmp	play
	
rx_empty:
	;empty message string: display "E"
	ldi	temp, 0x1C
	mov	col1, temp
	mov	col3, temp
	mov	col5, temp
	ldi	temp, 0x10
	mov	col2, temp
	mov	col4, temp
	rjmp	play_static
	
rx_checksum_error:
	;checksum error: display "CS"
	ldi	temp, 0x33
	mov	col1, temp
	ldi	temp, 0x44
	mov	col2, temp
	ldi	temp, 0x42
	mov	col3, temp
	ldi	temp, 0x41
	mov	col4, temp
	ldi	temp, 0x36
	mov	col5, temp
	rjmp	play_static
	
;--------------------
	
rx_read:
	;return current brightness in 'col1' (higher value = darker)
	sbi	PORTD, 3
	sbi	DDRD, 3
	rcall	muxwait
	cbi	DDRD, 3
	cbi	PORTD, 3
	ldi	bright, 0
rx_read_wait:
	sbis	PIND, 3
	rjmp	rx_read_done
	rcall	muxwait
	inc	bright
	cpi	bright, 255 ;timeout: limit time for very dark screens
	breq	PC+2
	rjmp	rx_read_wait
rx_read_done:
	ret
	
;===============================================================================

;---------------------------------------
;         RS232 TRANSFER MODE
;---------------------------------------

transfer:
	;init registers
	ldi	XL, RAM_String
	ldi	YL, 0x02
	
getc:
	;wait for transmission to begin
	sbis	PIND, 2
	rjmp	getc
	
	;receive byte
	rcall	waitbit ;startbit
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit
	
	rcall	waitbit ;data 0
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit
	
	rcall	waitbit ;data 1
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 2
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 3
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 4
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 5
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 6
	clc
	sbis	PIND, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 7
	clc
	sbis	PIND, 2
	sec
	ror	temp
	brcs	getc_starterr ;start bit was not 1
	rcall	waitbit

	rcall	waitbit
	sbic	PIND, 2
	rjmp	getc_stoperr ;stop bit is not 0
	
	;save byte to RAM buffer
	cpi	XL, RAM_String+112 ;prevent buffer overflow
	brlo	PC+2
	ldi	XL, RAM_String+111
	st	X+, temp

	;evaluate byte
	cpi	temp, 0x03 ;CTRL-C -> go to play mode
	breq	play
	cpi	temp, 0x04 ;CTRL-D -> evaluate string
	breq	getc_done
	cpi	temp, 0x0a ;LF -> evaluate string
	breq	getc_done
	cpi	temp, 0x0d ;CR -> evaluate string
	breq	getc_done
	cpi	temp, 0x08 ;BACKSP -> remove last character and the backspace character from buffer
	breq	getc_backsp
	rjmp	getc_end
	
getc_backsp:
	cpi	XL, RAM_String+2
	brlo	getc_end
	subi	XL, 2
	
getc_starterr:
getc_stoperr:
getc_end:
	;wait until line is idle (needed if stop bit is not 0) and receive next byte
	sbic	PIND, 2
	rjmp	getc_end
	rjmp	getc
	
;--------------------
	
getc_done:
	;terminate string
	dec	XL
	st	X, zero ;terminate string
getc_eval:
	;evaluate received string
	cpi	XL, RAM_String ;string is empty (contains only end marker)
	brne	PC+2
	rjmp	transfer
	ldi	XL, RAM_String
	ld	temp, X
	cpi	temp, 0x1b ;ESC
	breq	getc_control
	
	;save received string to EEPROM
	ldi	YL, 0x02
getc_savestr_loop:
	ld	temp, X+
	rcall	eeprom_write
	inc	YL
	cpi	temp, 0x00
	brne	getc_savestr_loop
	rjmp	transfer
	
getc_control:
	;control string received (begins with ESC)
	inc	XL
	ld	temp, X+
	cpi	temp, 's' ;scrolling speed
	breq	getc_speed
	cpi	temp, 'S'
	breq	getc_speed
	cpi	temp, 'q' ;quit -> switch to display mode
	breq	play
	cpi	temp, 'Q'
	breq	play
	cpi	temp, 0x1b ;ESC -> switch to display mode
	breq	play
	rjmp	transfer
	
getc_speed:
	;write scrolling speed setting to eeprom
	rcall	dec2bin
	ldi	YL, 0x01
	rcall	eeprom_write
	rjmp	transfer

;===============================================================================
	
;---------------------------------------
;             "PLAY" MODE
;---------------------------------------
	
play:
	;scrolling speed
	ldi	YL, 0x01
	rcall	eeprom_read
	mov	scrollspeed, temp
	
	;initialize registers
	ldi	YL, 0x02
	clr	YH
	ldi	charlen, 0
	mov	delay, scrollspeed
	mov	col1, zero
	mov	col2, zero
	mov	col3, zero
	mov	col4, zero
	mov	col5, zero
	mov	char1, zero
	mov	char2, zero
	mov	char3, zero
	mov	char4, zero
	mov	char5, zero
	
	;timer 0
	ldi	temp, 0x03 ;Clk/64
	out	TCCR0B, temp
	ldi	temp, 1<<TOIE0
	out	TIMSK, temp
	
	;enable interrupts
	sei
	
play_static:
	;ports
	ldi	temp, 0x03
	out	DDRA, temp
	ldi	temp, 0x04
	out	PORTA, temp
	ldi	temp, 0xFF
	out	DDRB, temp
	ldi	temp, 0x00
	out	PORTB, temp
	out	DDRD, zero
	out	PORTD, zero
	
loop:
	;multiplex display
	out	DDRD, col1
	ldi	temp, LINE1
	out	PORTB, temp
	rcall	muxwait
	ldi	temp, LINE_OFF
	out	PORTB, temp
	out	DDRD, zero
	
	out	DDRD, col2
	ldi	temp, LINE2
	out	PORTB, temp
	rcall	muxwait
	ldi	temp, LINE_OFF
	out	PORTB, temp
	out	DDRD, zero
	
	out	DDRD, col3
	ldi	temp, LINE3
	out	PORTB, temp
	rcall	muxwait
	ldi	temp, LINE_OFF
	out	PORTB, temp
	out	DDRD, zero
	
	out	DDRD, col4
	ldi	temp, LINE4
	out	PORTB, temp
	rcall	muxwait
	ldi	temp, LINE_OFF
	out	PORTB, temp
	out	DDRD, zero
	
	out	DDRD, col5
	ldi	temp, LINE5
	out	PORTA, temp
	rcall	muxwait
	ldi	temp, LINE5_OFF
	out	PORTA, temp
	out	DDRD, zero
	
	;check push-button
	ldi	temp, 0x04 ;PD4 = GND, pullup for PD2
	out	PORTD, temp
	ldi	temp, 0x10 ;PD4 = output, PD2 = input
	out	DDRD, temp
	rjmp	PC+1
	sbis	PIND, 2
	rjmp	rx
	out	DDRD, zero
	out	PORTD, zero
	
	rjmp	loop
	
;--------------------

dec2bin:
	;convert 3-digit decimal number from RAM[X,X+2] to 1 byte binary
	ld	temp2, X+ ;first digit: multiply by 100 = (64 + 32 + 4)
	andi	temp2, 0x03
	lsl	temp2 ;temp2 = temp2 * 2
	mov	temp, temp2
	swap	temp ;temp = (temp2*2) * 16 = temp2 * 32
	lsl	temp2 ;temp2 = (temp2*2) * 2 = temp2 * 4
	add	temp, temp2
	swap	temp2 ;temp2 = ((temp2*2)*2) * 16 = temp2 * 64
	add	temp, temp2
	
	ld	temp2, X+ ;second digit: multiply by 10 = (8 + 2)
	andi	temp2, 0x0F
	lsl	temp2 ;temp2 = temp2 * 2
	add	temp, temp2
	lsl	temp2 ;temp2 = (temp2*2) * 4 = temp2 * 8
	lsl	temp2
	add	temp, temp2
	
	ld	temp2, X+ ;first digit
	andi	temp2, 0x0F
	add	temp, temp2
	
	ret
	
;--------------------

muxwait:
	;wait a short time
	clr	w1
muxwait1:
	dec	w1
	brne	muxwait1
muxwait2:
	dec	w1
	brne	muxwait2
	ret

;--------------------

waitbit:
	;wait 1/2 rs232 bit length (203 cycles + rcall + 2 cycles for saving the bit)
	ldi	temp2, 66
waitbit_loop:
	dec	temp2
	brne	waitbit_loop
	nop
	
;	rjmp PC+1 ;slow down a little (funny controller or something ...)
;	rjmp PC+1
;	rjmp PC+1
;	rjmp PC+1
;	rjmp PC+1
	
	ret

;--------------------

eeprom_write:
	;write temp to EEPROM at ZL
	sbic	EECR, EEPE
	rjmp	eeprom_write
	out	EEAR, YL
	out	EEDR, temp
	sbi	EECR, EEMPE
	sbi	EECR, EEPE
	ret

;--------------------

eeprom_read:
	;read temp from EEPROM at YL
	sbic	EECR, EEPE
	rjmp	eeprom_read
	out	EEAR, YL
	sbi	EECR, EERE
	in	temp, EEDR
	ret
